/*
 * delveMatrix.c - Implementation of DELVE matrix facility.
 *
 * Copyright (c) 1996 by The University of Toronto.
 * 
 * See the file "copyright" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 * 
 * Author: Drew van Camp (drew@cs.toronto.edu)
 */

#ifndef lint
static char rcsid[] = "$Id: delveMatrix.c,v 1.3 1996/02/14 20:59:00 drew Exp $" ;
#endif


#include <stdio.h>
#include <tcl.h>
#include <tclExtend.h>
#include "delve.h"

static void	CleanUpMatrix _ANSI_ARGS_((void_pt matrixTablePtr,
			struct matrix_t *matrixPtr)) ;
static int	MatrixCreate _ANSI_ARGS_((Tcl_Interp *interp,
			void_pt matrixTablePtr, int rows, int cols)) ;
static int	MatrixDelete _ANSI_ARGS_((Tcl_Interp *interp,
			void_pt matrixTablePtr, char *handle)) ;
static int	MatrixSize _ANSI_ARGS_((Tcl_Interp *interp,
			void_pt matrixTablePtr, char *handle)) ;
static int	MatrixCopy _ANSI_ARGS_((Tcl_Interp *interp,
			void_pt matrixTablePtr, char *dstHandle, 
			char *srcHandle)) ;
static int	MatrixSet _ANSI_ARGS_((Tcl_Interp *interp,
			void_pt matrixTablePtr, char *handle, 
			char *valueString));
static int	MatrixSetEntry _ANSI_ARGS_((Tcl_Interp *interp,
			void_pt matrixTablePtr,char *handle, int row,
			int column, char *valueString)) ;
static int	MatrixGetEntry _ANSI_ARGS_((Tcl_Interp *interp,
			void_pt matrixTablePtr, char *handle, int row,
			int column)) ;
static int	MatrixCmd _ANSI_ARGS_((ClientData dummy, Tcl_Interp *interp,
			int argc, char **argv));
static void	MatrixCleanUp _ANSI_ARGS_((ClientData clientData,
			Tcl_Interp *interp)) ;


/*
 *-----------------------------------------------------------------------------
 * CleanUpMatrix --
 *
 *   Release all resources allocated to the specified matrix.  Doesn't
 * free the table entry.
 *-----------------------------------------------------------------------------
 */
static void
CleanUpMatrix (matrixTablePtr, matrixPtr)
void_pt		matrixTablePtr;
struct matrix_t *matrixPtr;
{
    int		idx ;

    for (idx = 0 ; idx < matrixPtr->rows ; ++idx) {
	ckfree((char *)matrixPtr->x[idx]) ;
    }
    ckfree((char *)matrixPtr->x) ;
    ckfree((char *)matrixPtr) ;
}

/*
 *-----------------------------------------------------------------------------
 * MatrixCreate --
 *
 *   Create a new matrix, implements the subcommand:
 *         matrix create rows cols
 *-----------------------------------------------------------------------------
 */
static int
MatrixCreate (interp, matrixTablePtr, rows, cols)
Tcl_Interp *	interp;
void_pt      	matrixTablePtr;
int		rows ;
int		cols ;
{
    struct matrix_t *	matrixPtr;
    struct matrix_t **	tableEntryPtr;
    double **	x ;
    int		i ;
    int		j ;

    /*
     * Make sure that the row and column sizes are in the right range.
     */

    if (rows <= 0) {
	char	buffer[200] ;
	sprintf(buffer, "expected row size greater than 0, but got %d", 
		rows) ;
	Tcl_SetResult(interp, buffer, TCL_VOLATILE) ;
	return TCL_ERROR ;
    }

    if (cols <= 0) {
	char	buffer[200] ;
	sprintf(buffer, "expected column size greater than 0, but got %d", 
		cols) ;
	Tcl_SetResult(interp, buffer, TCL_VOLATILE) ;
	return TCL_ERROR ;
    }

    /*  
     * Allocate all the memory, and set everything to zero.
     */

    matrixPtr = (struct matrix_t *) ckalloc (sizeof (struct matrix_t));

    x = (double **)ckalloc(rows*sizeof(double *)) ;
    for (i = 0 ; i < rows ; ++i) {
	x[i] = (double *)ckalloc(cols*sizeof(double)) ;
	for (j = 0 ; j < cols ; ++j) {
	    x[i][j] = 0.0 ;
	}
    }

    matrixPtr->x    = x ;
    matrixPtr->rows = rows ;
    matrixPtr->cols = cols ;

    /*
     * Add the matrix into the handle table, and return the handle in
     * the interpreters result.
     */

    tableEntryPtr = (struct matrix_t **)
        Tcl_HandleAlloc (matrixTablePtr, matrixPtr->handle);
    *tableEntryPtr = matrixPtr;

    Tcl_SetResult (interp, matrixPtr->handle, TCL_STATIC);
    return TCL_OK;
}

/*
 *-----------------------------------------------------------------------------
 * MatrixDelete --
 *
 *   Deletes the specified matrix, implements the subcommand:
 *         matrix delete handle
 *-----------------------------------------------------------------------------
 */
static int
MatrixDelete (interp, matrixTablePtr, handle)
Tcl_Interp *	interp;
void_pt		matrixTablePtr;
char *		handle;
{
    struct matrix_t **	tableEntryPtr;
    double **	x ;

    tableEntryPtr
	= (struct matrix_t **)Tcl_HandleXlate (interp, matrixTablePtr, handle);

    if (tableEntryPtr == NULL) {
        return TCL_ERROR;
    }

    CleanUpMatrix(matrixTablePtr, *tableEntryPtr) ;

    Tcl_HandleFree (matrixTablePtr, tableEntryPtr);

    return TCL_OK;
}


/*
 *-----------------------------------------------------------------------------
 * MatrixSize --
 *
 *   Prints the number of rows and columns in the matrix into the
 *   interpreters of the result. Implements the subcommand:
 *	matrix size
 *
 *-----------------------------------------------------------------------------
 */
static int
MatrixSize (interp, matrixTablePtr, handle)
Tcl_Interp *	interp;
void_pt		matrixTablePtr;
char *		handle;
{
    matrix_t **	tableEntryPtr;
    matrix_t *	matrixPtr;

    tableEntryPtr
	= (matrix_t **) Tcl_HandleXlate (interp, matrixTablePtr, handle);

    if (tableEntryPtr == NULL) {
        return TCL_ERROR;
    }

    matrixPtr = *tableEntryPtr ;
    
    sprintf(interp->result, "%d %d", matrixPtr->rows, matrixPtr->cols) ;

    return TCL_OK;
}


/*
 *-----------------------------------------------------------------------------
 * MatrixCopy --
 *
 *   Copies the values in one matrix to another. Implements the
 *   subcommand:
 *	matrix copy dstHandle srcHandle
 *
 *-----------------------------------------------------------------------------
 */
static int
MatrixCopy (interp, matrixTablePtr, dstHandle, srcHandle)
Tcl_Interp *	interp;
void_pt		matrixTablePtr;
char *		srcHandle;
char *		dstHandle;
{
    int		i ;
    int		j ;
    matrix_t *	matrixPtr[2];
    matrix_t **	tableEntryPtr;

    tableEntryPtr
	= (matrix_t **) Tcl_HandleXlate (interp, 
					       matrixTablePtr, dstHandle);

    if (tableEntryPtr == NULL) {
        return TCL_ERROR;
    }

    matrixPtr[0] = *tableEntryPtr ;

    tableEntryPtr
	= (matrix_t **) Tcl_HandleXlate (interp, 
					       matrixTablePtr, srcHandle);

    if (tableEntryPtr == NULL) {
        return TCL_ERROR;
    }

    matrixPtr[1] = *tableEntryPtr ;

    if (matrixPtr[1]->rows != matrixPtr[0]->rows) {
	Tcl_AppendResult(interp, "matrices have different number of rows",
			 (char *) NULL) ;
	return TCL_ERROR;
    }

    if (matrixPtr[1]->cols != matrixPtr[0]->cols) {
	Tcl_AppendResult(interp, "matrices have different number of columns",
			 (char *) NULL) ;
	return TCL_ERROR;
    }

    for (i = 0 ; i < matrixPtr[0]->rows ; ++i) {
	for (j = 0 ; j < matrixPtr[0]->cols ; ++j) {
	    matrixPtr[0]->x[i][j] = matrixPtr[1]->x[i][j] ;
	}
    }

    return TCL_OK;
}


/*
 *-----------------------------------------------------------------------------
 * MatrixSet --
 *
 *   Sets all elements in the matrix to a single value. Implements the
 *   subcommand:
 *	matrix set value
 *
 *-----------------------------------------------------------------------------
 */
static int
MatrixSet (interp, matrixTablePtr, handle, valueString)
Tcl_Interp *	interp;
void_pt		matrixTablePtr;
char *		handle;
char *		valueString ;
{
    double	value ;
    int		i ;
    int		j ;
    matrix_t *	matrixPtr;
    matrix_t **	tableEntryPtr;

    tableEntryPtr
	= (matrix_t **) Tcl_HandleXlate (interp, matrixTablePtr, handle);

    if (tableEntryPtr == NULL) {
        return TCL_ERROR;
    }

    matrixPtr = *tableEntryPtr ;

    if (Tcl_GetDouble(interp, valueString, &value) != TCL_OK) {
	Tcl_AddErrorInfo(interp,
		 "\n    (while parsing value to set all matrix elements to)") ;
	return TCL_ERROR ;
    }

    for (i = 0 ; i < matrixPtr->rows ; ++i) {
	for (j = 0 ; j < matrixPtr->cols ; ++j) {
	    matrixPtr->x[i][j] = value ;
	}
    }

    return TCL_OK;
}


/*
 *-----------------------------------------------------------------------------
 * MatrixSetEntry --
 *
 *   Set an entry in the matrix from a string. Implements the subcommand:
 * 	matrix entryset i j val
 *-----------------------------------------------------------------------------
 */
static int
MatrixSetEntry (interp, matrixTablePtr, handle, row, col, valueString)
Tcl_Interp *	interp;
void_pt		matrixTablePtr;
char *		handle;
int		row ;
int		col ;
char *		valueString ;
{
    char	buffer[200] ;
    double	value ;
    matrix_t *	matrixPtr;
    matrix_t **	tableEntryPtr;

    tableEntryPtr
	= (matrix_t **) Tcl_HandleXlate (interp, matrixTablePtr, handle);

    if (tableEntryPtr == NULL) {
        return TCL_ERROR;
    }

    matrixPtr	= *tableEntryPtr;

    /*
     * First make sure that the row and column indices are in the
     * right range.
     */

    if (row < 0 || row >= matrixPtr->rows) {
	sprintf(buffer, "expected row index between 0 and %d, but got %d", 
		matrixPtr->rows - 1, row) ;
	Tcl_SetResult(interp, buffer, TCL_VOLATILE) ;
	Tcl_AddErrorInfo(interp, "\n    (while setting element for matrix)") ;
	return TCL_ERROR ;
    }

    if (col < 0 || col >= matrixPtr->cols) {
	sprintf(buffer, "expected column index between 0 and %d, but got %d", 
		matrixPtr->cols - 1, col) ;
	Tcl_SetResult(interp, buffer, TCL_VOLATILE) ;
	Tcl_AddErrorInfo(interp, "\n    (while setting element for matrix)") ;
	return TCL_ERROR ;
    }

    /*
     * Parse the value to set the element to, and set it.
     */

    if (Tcl_GetDouble(interp, valueString, &value) != TCL_OK) {
	sprintf(buffer,
		"\n    (while parsing element for matrix entry %d,%d)",
		row, col) ;
	Tcl_AddErrorInfo(interp, buffer) ;
	return TCL_ERROR ;
    }

    matrixPtr->x[row][col] = value ;

    return TCL_OK;
}


/*
 *-----------------------------------------------------------------------------
 * MatrixGetEntry --
 *
 *   Returns a string representation of the value of matrix[row][col]
 *   in the interpreters result. 
 *   Implements the subcommand: 
 * 	matrix entryset i j 
 *-----------------------------------------------------------------------------
 */

static int
MatrixGetEntry (interp, matrixTablePtr, handle, row, col)
Tcl_Interp *	interp;
void_pt		matrixTablePtr;
char *		handle;
int		row ;
int		col ;
{
    char	buffer[200] ;
    char	dblBuf[TCL_DOUBLE_SPACE] ;
    matrix_t *	matrixPtr;
    matrix_t **	tableEntryPtr;

    tableEntryPtr
	= (matrix_t **) Tcl_HandleXlate (interp, matrixTablePtr, handle);

    if (tableEntryPtr == NULL) {
        return TCL_ERROR;
    }

    matrixPtr = *tableEntryPtr;

    /*
     * First make sure that the row and column indices are in the
     * right range.
     */

    if (row < 0 || row >= matrixPtr->rows) {
	sprintf(buffer, "expected row index between 0 and %d, but got %d", 
		matrixPtr->rows - 1, row) ;
	Tcl_SetResult(interp, buffer, TCL_VOLATILE) ;
	Tcl_AddErrorInfo(interp, "\n    (while setting element for matrix)") ;
	return TCL_ERROR ;
    }

    if (col < 0 || col >= matrixPtr->cols) {
	sprintf(buffer, "expected column index between 0 and %d, but got %d", 
		matrixPtr->cols - 1, col) ;
	Tcl_SetResult(interp, buffer, TCL_VOLATILE) ;
	Tcl_AddErrorInfo(interp, "\n    (while setting element for matrix)") ;
	return TCL_ERROR ;
    }

    /*
     * Print the value into the interpreter's result, and return ;
     */

    Tcl_PrintDouble(interp, matrixPtr->x[row][col], dblBuf) ;
    Tcl_SetResult(interp, dblBuf, TCL_VOLATILE) ;

    return TCL_OK;
}


/*
 *-----------------------------------------------------------------------------
 * MatrixCmd --
 *
 *    The procedure for the "matrix" command.
 *-----------------------------------------------------------------------------
 */
static int
MatrixCmd(clientData, interp, argc, argv)
ClientData 	clientData;	/* Handle table pointer. */
Tcl_Interp *	interp;		/* Current interpreter. */
int 		argc;		/* Number of arguments. */
char **		argv;		/* Argument strings. */
{
    char *	option ;
    char	c ;

    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # args: should be \"",
			 argv[0], " option ?arg ...?\"", (char *) NULL) ;
	return TCL_ERROR;
    }

    option = argv[1] ;
    c = option[0] ;

    if (c == 'c' && strcmp(option, "copy") == 0) {
	if (argc != 4) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
			     argv[0],
			     " copy destMatrixHandle srcMatrixHandle\"", 
			     (char *) NULL);
	    return TCL_ERROR;
	}
	if (MatrixCopy(interp, (void_pt) clientData, 
		       argv[2], argv[3]) != TCL_OK) {
	    return TCL_ERROR ;
	}

    } else if (c == 'c' && strcmp(option, "create") == 0) {
	int	rows ;
	int	cols ;

	if (argc != 4) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
			     argv[0], " create rows columns\"", (char *) NULL);
	    return TCL_ERROR;
	}
	if (Tcl_GetInt(interp, argv[2], &rows) != TCL_OK) {
	    return TCL_ERROR ;
	}
	if (Tcl_GetInt(interp, argv[3], &cols) != TCL_OK) {
	    return TCL_ERROR ;
	}
	if (MatrixCreate(interp, (void_pt) clientData, rows, cols) != TCL_OK) {
	    return TCL_ERROR ;
	}

    } else if (c == 'd' && strcmp(option, "delete") == 0) {
	if (argc != 3) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
			     argv[0], " delete matrixHandle\"", (char *) NULL);
	    return TCL_ERROR;
	}
	if (MatrixDelete(interp, (void_pt) clientData, argv[2]) != TCL_OK) {
	    return TCL_ERROR ;
	}
	
    } else if (c == 'e' && strcmp(option, "entryset") == 0) {
	int	 row ;
	int	 column ;

	if (argc != 5 && argc != 6) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
			     argv[0], 
			     " entryset matrixHandle row column ?value?\"", 
			     (char *) NULL) ;
	    return TCL_ERROR;
	}

	if (Tcl_GetInt(interp, argv[3], &row) != TCL_OK) {
	    return TCL_ERROR ;
	}
	if (Tcl_GetInt(interp, argv[4], &column) != TCL_OK) {
	    return TCL_ERROR ;
	}
	if (argc == 6 
	    && MatrixSetEntry(interp, (void_pt) clientData, 
			      argv[2], row, column, argv[5]) != TCL_OK) {
	    return TCL_ERROR ;
	}
	if (MatrixGetEntry(interp, (void_pt) clientData, 
			    argv[2], row, column) != TCL_OK) {
	    return TCL_ERROR ;
	}
	
    } else if (c == 's' && strcmp(option, "set") == 0) {
	if (argc != 4) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
			     argv[0], " set matrixHandle value\"",
			     (char *) NULL) ;
	    return TCL_ERROR;
	}
	if (MatrixSet(interp, (void_pt) clientData, 
		      argv[2], argv[3]) != TCL_OK) {
	    return TCL_ERROR ;
	}
	
    } else if (c == 's' && strcmp(option, "size") == 0) {
	if (argc != 3) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
			     argv[0], " size matrixHandle\"", (char *) NULL);
	    return TCL_ERROR;
	}
	if (MatrixSize(interp, (void_pt) clientData, argv[2]) != TCL_OK) {
	    return TCL_ERROR ;
	}
	
    } else {
	Tcl_AppendResult(interp, "bad option \"",
			 option, "\": should be copy, create, ",
			 "delete, entryset, set, or size", (char *) NULL) ;
	return TCL_ERROR;
    }

    return TCL_OK ;
}

/*
 *-----------------------------------------------------------------------------
 * MatrixCleanUp --
 *
 *    Called when the interpreter is deleted to cleanup all matrix
 * resources
 *-----------------------------------------------------------------------------
 */
static void
MatrixCleanUp (clientData, interp)
ClientData	clientData ;
Tcl_Interp * 	interp;
{
    struct matrix_t **	tableEntryPtr;
    int		walkKey;
    
    walkKey = -1;
    while (1) {
        tableEntryPtr =
            (struct matrix_t **)Tcl_HandleWalk((void_pt) clientData, &walkKey);
        if (tableEntryPtr == NULL) {
            break;
	}
        CleanUpMatrix ((void_pt) clientData, *tableEntryPtr);
    }
    Tcl_HandleTblRelease ((void_pt) clientData);
}

/*
 *-----------------------------------------------------------------------------
 *  Delve_InitMatrix --
 *
 *    Initialize the DELVE matrix facility.
 *-----------------------------------------------------------------------------
 */
int
Delve_InitMatrix (interp)
    Tcl_Interp *interp;
{
    void_pt  matrixTablePtr;

    matrixTablePtr 
	= Tcl_HandleTblInit ("matrix", sizeof (struct matrix_t *), 10);

    Tcl_CallWhenDeleted (interp, MatrixCleanUp, (ClientData) matrixTablePtr);

    /*
     * Initialize the commands.
     */

    Tcl_CreateCommand (interp, "d_matrix", MatrixCmd, 
                       (ClientData) matrixTablePtr, (void (*)()) NULL);

    return TCL_OK ;
}

/*
 *-----------------------------------------------------------------------------
 *  Delve_GetMatrix --
 *
 *    Finds a matrix given it's handle
 *-----------------------------------------------------------------------------
 */
int
Delve_GetMatrix (interp, handle, matrixPtrPtr)
Tcl_Interp *	interp ;
CONST char *	handle ;
matrix_t **matrixPtrPtr ;
{
    matrix_t **	tableEntryPtr;
    void_pt	matrixTablePtr;
    Tcl_CmdInfo	info ;

    if (Tcl_GetCommandInfo(interp, "d_matrix", &info) == 0
	|| info.proc != MatrixCmd) {
	Tcl_SetResult(interp, "can't find \"d_matrix\" command in interpreter",
		      TCL_VOLATILE) ;
	return TCL_ERROR ;
    }
    
    matrixTablePtr = (void_pt) info.clientData ;
    tableEntryPtr
	= (matrix_t **) Tcl_HandleXlate (interp, matrixTablePtr, handle);

    if (tableEntryPtr == NULL) {
        return TCL_ERROR;
    }
    *matrixPtrPtr = *tableEntryPtr ;

    return TCL_OK ;
}
