% ----------------------------------------------------------------------
%
% Tcl utilities for handling tasks
%
% Author: Drew van Camp (drew@@cs.toronto.edu)
%
% Copyright (c) 1995-1996 The University of Toronto.
%
% See the file "copyright" for information on usage and redistribution
% of this file, and for a DISCLAIMER OF ALL WARRANTIES.
%
% ----------------------------------------------------------------------

\documentclass[dvips]{article}
\usepackage{moretext,amstex,alltt,changebar}
\usepackage{delve}
\begin{document}

\rcsInfo $Id: taskGen.w,v 1.22 1996/05/14 14:24:05 drew Exp $

\title{TaskGen: Generation of Delve Task Files}
\author{Drew van Camp (drew@@@@cs.utoronto.ca)\\[1ex]
	Department of Computer Science\\
	University of Toronto\\
	6 Kings College Road\\
	Toronto ON, Canada, M5S 1A4}

\vfil
\maketitle
\vfil
\copyrightNotice{1995-1996}
\vfil
\clearpage
\tableofcontents
\clearpage

\section{Introduction}
This module defines utilities that are used for manipulating tasks and
obtaining information about them. It uses routines from the ``Utils''
module. 

\section{Saving encoding schemes: \texttt{saveEncoding}} 
\label{sec:saveEncoding}
The \texttt{saveEncoding} command is used to save the encoding scheme
that is currently in use to a file.  The scheme can later be restored
from the file by calling \texttt{loadEncoding}.
\subsection{Description}
\usage{saveEncoding}{file ?force?}
This procedure writes to a file the encoding scheme for all attributes
registered for encoding. The format of the \texttt{file} is similar to
a prior information file. Each line has the form:
\begin{alltt}
attr method ?option=value ...?
\end{alltt}
\texttt{attr} can be either the index number of an attribute, or a
symbolic name. \texttt{method} is the encoding method to be used (see
the \texttt{acdc} module for a description of these). If the encoding
method has options, then they should appear later on the line with the
name of the option \textit{no space}, an equals sign \textit{no
space}, and the value for the option. Any number of option/value pairs
may occur on a line. By default, the procedure will not over-write
existing files; however if the optional \texttt{force} is given and is
non-zero, it will silently over-write them.

\subsection{Implementation}
The procedure has one required argument, the name of the file to
write, and one optional argument \texttt{force}, for controlling the
over-writing of pre-existing files. It first opens the output file. It
then iterates over all attributes, massaging the the attribute name,
encoding method, and options into the proper format, and printing them
to the file. Finally, it closes the file.

@o taskGen.tcl @{
proc saveEncoding { file { force 0 } } {
    @<open the output file for the encoding@>
    foreach encoding [acdc_list] {
	@<format the attribute name@>
	@<format the encoding method@>
	@<format the encoding options@>
	@<print the encoding to the file@>
    }
    @<close the output file for the encoding@>
}
@}

Opening and closing the output file are straight-forward; however we
must be sure to use the proper opening mode based on whether or not
\texttt{force} is true. We use \texttt{wopen} for this.

@d open the output file for the encoding @{
set out	 [wopen $file $force] ;
@}

@d close the output file for the encoding @{
close $out ;
@}

The name of the attribute is the \texttt{id}, or second element of the
encoding list returned by \texttt{acdc\_list}:
@d format the attribute name @{
set name	[lindex $encoding 1] ;
@}

The encoding method can be obtained using \texttt{acdc\_method}
@d format the encoding method @{
set method	[acdc_method $name] ;
@}

The only options that can be specified in an encoding file are:
\texttt{passive}, \texttt{center}, and \texttt{unit}. The values for
these options can be obtained using \texttt{acdc\_options}, and
formatted separately if, and only if, they have been specified.

@d format the encoding options @{
set options {} ;
foreach option [list passive center unit] {
    set value [acdc_options $name -$option] ;
    if ![cequal $value {}] {
	append options " $option=[list $value]"
    }
}
@}

Finally, we just have to print all the info out in the proper format.
@d print the encoding to the file @{
puts $out "[list $name $method] $options" ;
@}

\section{Restoring an encoding scheme: \texttt{loadEncoding}}
The \texttt{loadEncoding} command is used to save the restore an
encoding scheme that was previously saved to a file by
\texttt{loadEncoding}.  
\subsection{Description}
\usage{loadEncoding}{file attrs}
This procedure loads an encoding scheme for a set of attributes. See
Section~\ref{sec:saveEncoding} for a description of the file format.

All attributes mentioned in \texttt{file} must also appear in the list
\texttt{attrs} (i.e. must have already been registered for encoding).

\subsection{Implementation}
The procedure has two required arguments, the name of the file to
load, and the list of allowed attributes.  It iterates over all lines
in the file by reading the file in all at once using
\texttt{read\_file}, then splitting the result up into a list, where
each line becomes an element.

@o taskGen.tcl @{
proc loadEncoding { file attrs } {
    foreach line [split [read_file -nonewline $file] "\n"] {
	@<find the attribute name@>
	@<set the attribute encoding method@>
	@<set any options@>
    }
}
@}

When we get the attribute's name, we must also check if it has been
previously registered with the encoding module, and is in the list
\texttt{attrs}. We use the \texttt{acdc\_method} procedure for this.

@d find the attribute name @{
set attr 	[lvarpop line] ;
if [catch {acdc_list $attr} list] {
    error "attribute \"$attr\" doesn't exist (file $file)" ;
}
set aliases	[lindex $list 0] ;
if [lempty [union [concat [list $attr] $aliases] $attrs]] {
    error "attribute \"$attr\" exists, but isn't valid (file $file)" ;
}

@}

Setting the attribute's encoding scheme is straight-forward. The only
trick is passing the \texttt{--} argument to \texttt{acdc\_method} in
case \texttt{method} starts with a dash.

@d set the attribute encoding method @{
set method 	[lvarpop line] ;
acdc_method $attr -- $method
@}

When setting options, we first clear any previous values, then iterate
over all remaining arguments on the line splitting them up at the
equals sign (\texttt{=}). We make sure that we don't care about case
by converting the key to all lower case letters.

@D set any options @{
foreach key [list passive center unit] {
    acdc_options $attr -$key {} ;
}

foreach option $line {
    if ![regexp {^([^=]*)=([^=]*)$} $option all key value] {
	error "expected \"key=value\" but got \"$option\" (file $file)"
    }
    set key	[string tolower $key] ;
    acdc_options $attr -$key $value ;
}
@}


\section{Generating Task Instances: \texttt{genTask}, \texttt{genTasInstances}}
These procedures are used to generate the data files necessary to run
on one or more task instances. 
\subsection{Description}
\usage{genTaskInstances}{cpath path fileList instList force}
The procedure \texttt{genTaskInstances} is used to generate the data
files necessary to run on one or more task instances. This includes
for each requested instance a \texttt{training} file, a
\texttt{testing} input file, a testing \texttt{target} file, and a
\texttt{normalize} file. As well, a single encoding file, and a single
test set statistics file are generated. Arguments to the procedure are:
\begin{description}
\item[\texttt{cpath}] - the canonical name of the task being generated.
\item[\texttt{path}] - the path of the directory in which the files
	are to be created.
\item[\texttt{fileList}] - this list should contain at least one of
	\texttt{training}, \texttt{testing}, and \texttt{targets}. The
	procedure will generate only the files in the list. The
	\texttt{normalize} file is always generated.
\item[\texttt{instList}] - a list of integer indices of the instances
	you want to generate, or the string \texttt{all} to generate
	all instances.
%\item[\texttt{scheme}] - the scheme to be used when selecting testing
%	data. It should be one of: \texttt{common} to create test sets
%	that all share the same data; or \texttt{hierarchical} to
%	create test sets where each instance uses unique subset of the
%	testing data. \texttt{Common} test sets will contain
%	\texttt{n} times as many cases as \texttt{hierarchical} test
%	sets, where \texttt{n} is the total number of instances in the
%	task specified by \texttt{cpath}.
\item[\texttt{force}] - a boolean value. If it is 1, existing files
	will be silently overwritten. If it is 0, an error will be
	generated if any file already exists. 
\end{description}
\usage{genTask}{cpath path scheme force}
This procedure is a simple front end to \texttt{genTask}. It is
equivalent to:
\begin{alltt}
genTaskInstances cpath path "training testing targets" all force
\end{alltt}

\subsection{Implementation}
The procedure implementing \texttt{genTaskInstances} first segments
the prototask file into raw training, testing, and target files. If
the \texttt{instList} contains \texttt{all}, it sets it to a list of
all the possible instances.  Next it iterates over all of the
instances encoding the appropriate raw files from the
\texttt{fileList} and creating the \texttt{normalize} file. Next, it
calculates statistics based on the raw target file. Finally, it
deletes all temporary files.

@o taskGen.tcl @{
proc genTaskInstances { cpath path fileList instList force } {
    @<find the prototask file@>
    @<split up prototask file into training and test files@>
    @<encode the files for each instance@>
    @<calculate the test target stats@>
    @<save the encoding information@>
    @<delete all the raw test and training files@>
}
@}

\subsubsection{Finding the prototask file}

Before we can do anything, we have to find the file containing all of
the prototask data. The name of the file is stored in the global
variable \texttt{d\_priv(prototaskFile)}. The path of the directory
holding the file is simply the directory part of the tasks canonical
name.  Once we've created the full canonical name of the file, we can
pass the name to \texttt{d\_getFile} which will pass in the full path
name of the file, or generate an error if it doesn't exist, or there
are multiple copies.

@d find the prototask file @{
global d_priv ;
set tail	  $d_priv(prototaskFile) ;
set prototaskFile [d_getFile [file dirname $cpath]/$tail "data file"] ;
@}

\subsubsection{Splitting up the prototask file}

To generate the task instances, we have to segment the prototask file,
putting the appropriate cases into raw case files: one containing
testing inputs, one containing testing targets, and several containing
both training inputs and targets.

To do this, we call a procedure \texttt{splitPrototask} described in
Section~\ref{sec:splitPrototask}. We have to tell the procedure the
name of the file to split, the number of input attributes, the number
of target attributes, the number of cases in the test set, 
\begin{changebar}
the number of cases in each training set, and the maximum number of
instances to generate.  
\end{changebar}
\begin{changebar}
As well, we have to pass the scheme to be used when selecting testing
data. It should be one of: \texttt{common} to create test sets that
all share the same data; or \texttt{hierarchical} to create test sets
where each instance uses unique subset of the testing
data. \texttt{Common} test sets will contain \texttt{n} times as many
cases as \texttt{hierarchical} test sets, where \texttt{n} is the
total number of instances in the task specified by \texttt{cpath}.
\texttt{SplitPrototask} then returns three lists The lists contain:
the names of the raw test input files, the names of the raw test
target files, and the names of the raw training files.
\end{changebar}
The files it generates are written into the
\texttt{/tmp/} directory, so we don't have to worry about over-writing
anything.

\begin{changebar}
@d split up prototask file into training and test files @{
set files [splitPrototask $prototaskFile \
	       [llength [d_info $cpath inputs]] \
	       [llength [d_info $cpath targets]] \
	       [d_info $cpath test-set-size] \
	       [d_info $cpath training-set-size] \
	       [d_info $cpath maximum-number-of-instances] \
	       [d_info $cpath test-set-selection]] ;
set rawFiles(test)	[lvarpop files] ;
set rawFiles(targets)	[lvarpop files] ;
set rawFiles(training)	[lvarpop files] ;
@}
\end{changebar}

\subsubsection{Encode the instance files}

Now that all the raw files have been generated, we have to iterate
over each instance, calculating the raw training set statistics that
will be used to encode everything, then generating the appropriate
files based on the list in the argument \texttt{fileList}.

However, before we can do that, we have to find the list of instances
to iterate over: if the argument \texttt{instList} contained the
string \texttt{all}, then we have to count all the possible test
instances and set \texttt{instList} to them; otherwise we can just use
its value.

\begin{changebar}
@D encode the files for each instance @{
if [cequal $instList "all"] {
    set instList {} ;
    loop idx 0 [llength $rawFiles(training)] {
	lappend instList $idx ;
    }
}

foreach instance $instList {
    set rawTrainingFile [lindex $rawFiles(training) $instance] ;

    @<save the raw training set stats to the normalize file@>

    set regexp \
	{^normalization options ((mean)|(stdDev)|(median)|(absDev)) (.*) (.*)}
    for_file line $path/normalize.$instance {
	if [regsub $regexp $line {acdc_options \6 -\1 \7} new] {
	    eval $new ; 
	} else {
	    error "unknown line in $path/normalize.$instance: \"$line\"" ;
	}
    }

    if { [lsearch $fileList "training"] >= 0 } {
	pwrap "  encoding instance $instance training data..." {
	    set fromFileId	[open $rawTrainingFile r] ;
	    set toFileId	[wopen $path/train.$instance $force] ;
	    acdc_encodeFile $attrList $fromFileId $toFileId ;
	    close $fromFileId ;
	    close $toFileId ;
	}
    }
    if { [lsearch $fileList "testing"] >= 0 } {
	pwrap "  encoding instance $instance test inputs..." {
	    set fromFileId	[open [lindex $rawFiles(test) $instance] r] ;
	    set toFileId	[wopen $path/test.$instance $force] ;
	    acdc_encodeFile [d_info $cpath inputs] $fromFileId $toFileId ;
	    close $fromFileId ;
	    close $toFileId ;
	}
    }
    if { [lsearch $fileList "targets"] >= 0 } {
	pwrap "  encoding instance $instance test targets..." {
	    set fromFileId	[open [lindex $rawFiles(targets) $instance] r]
	    set toFileId	[wopen $path/targets.$instance $force] ;
	    acdc_encodeFile [d_info $cpath targets] $fromFileId $toFileId ;
	    close $fromFileId ;
	    close $toFileId ;
	}
    }
}
@}
\end{changebar}

To encode the raw data for a particular instance, we need the mean,
median, variance, and average absolute deviation for each attribute in
the training data for the instance.  To obtain these we use the \delve\
\texttt{d\_matrix} command.  

We first load all of the \texttt{real} and \texttt{integer} attribute
values from the raw training data file into a matrix.  Once we've done
that we can iterate over the same attributes, and save their,
statistcist to the file \texttt{normalize.\$instance}.  The statistics
are calculated by calling the \texttt{d\_mstats} command which returns
a list containing the mean, variance, median and average absolute
deviation from the median for a particular row or column of the
matrix.

We use the \texttt{force} option to determine the mode for opening the
file, and save the values in a form that can be simply sourced to
reload the values. 

@D save the raw training set stats to the normalize file @{
set attrList	[concat [d_info $cpath inputs] [d_info $cpath targets]] ;
set indexList	{} ;
set index 	0 ;
foreach list [eval acdc_list $attrList] {
    set type [lindex $list 0] ;
    if { [cequal $type real] || [cequal $type integer] } {
	lappend indexList $index ;
    }
    incr index ;
}
set matrix \
    [d_matrix create [d_info $cpath training-set-size] [llength $indexList]] ;
loadMatrix $matrix $rawTrainingFile $indexList ;

set out	      [wopen $path/normalize.$instance $force] ;

set col 0 ;
foreach index $indexList {
    set attr [lindex $attrList $index] ;
    set list [d_mstats -col $col $matrix] ;

    puts $out [list normalization options mean $attr [lvarpop list]] ;
    puts $out [list normalization options stdDev $attr [sqrt [lvarpop list]]] ;
    puts $out [list normalization options median $attr [lvarpop list]] ;
    puts $out [list normalization options absDev $attr [lvarpop list]] ;

    incr col ;
}

d_matrix delete $matrix ;
close $out ;
@}

\subsubsection{Test set statistics} 
The raw testing set target value statistics are calculated in the same
manner as the raw training set statistics. Since these statistics are
used by \texttt{deval}, we have to dump them to a file. For now we'll
call it \texttt{Test-set-stats}.  The first line will contain a list of
the variances for all \texttt{real} and \texttt{integer} targets. The
second line, their average absolute deviations from the median.
If the test set selection scheme is \texttt{common}, all target files
are identical, and we only need to calculate the statistics using one
of the files.  If the scheme is \texttt{hierarchical}, we must
concatenate all of the files into a single file, then calculate
statistics using it.
@d get the file to calculate test target stats from @{
set scheme	[d_info $cpath test-set-selection] ;
if { [cequal $scheme "common"] } {
    set targetFile [lindex $rawFiles(targets) 0] ;
} elseif { [cequal $scheme "hierarchical"] } {
    set targetFile "/tmp/alltargets.[pid]" ;
    set out	   [open $targetFile w] ;
    foreach infile $rawFiles(targets) {
	set in [open $infile r] ;
	copyfile $in $out ;
	close $in ;
    }
    close $out ;
} else {
    error "unknown test set selection scheme \"$scheme\"" ;
}
@}

We can only calculate the statistics if there is at least one
\texttt{real} or \texttt{integer} target variable.  This necessitates
a very large \texttt{if} branch below.

@D calculate the test target stats @{
set attrList	[d_info $cpath targets] ;
set indexList	{} ;
set index 	0 ;
foreach list [eval acdc_list $attrList] {
    set type [lindex $list 0] ;
    if { [cequal $type real] || [cequal $type integer] } {
	lappend indexList $index ;
    }
    incr index ;
}

if { ![lempty $indexList] } {
    set matrix \
	[d_matrix create [d_info $cpath test-set-size] [llength $indexList]] ;

    @<get the file to calculate test target stats from@>
    loadMatrix $matrix $targetFile $indexList ;
    if { [cequal $scheme "hierarchical"] } {
	unlink -nocomplain $targetFile ;
    }
} elseif { [info exists matrix] } {
    unset matrix ;
}
    
set col	0 ;
set variances	{} ;
set absDevs	{} ;
foreach index $indexList {
    set attr [lindex $attrList $index] ;
    set list [d_mstats -col $col $matrix] ;
    lappend variances	[lindex $list 1] ;
    lappend absDevs	[lindex $list 3] ;
    incr col ;
}

set out	[wopen $path/Test-set-stats $force] ;
if [info exists matrix] {
    d_matrix delete $matrix ;
    puts $out "$variances\n$absDevs"
}
close $out ;
@}

\subsubsection{Save the encoding information}
The encoding information is written to the file
\texttt{Coding-used}. It can be loaded later to be used by other
scripts that need to know how the data was encoded.

@d save the encoding information @{
saveEncoding $path/Coding-used $force ;
@}

\subsubsection{Cleanup}

The final task to perform in \texttt{genTaskInstances} is the cleanup:
we simply delete the temporary files.

@d delete all the raw test and training files @{
foreach key [array names rawFiles] {
    unlink -nocomplain $rawFiles($key) ;
}
@}

The implementation of \texttt{genTask}is very straight-forward:
@o taskGen.tcl @{
proc genTask { cpath path { force 0 } } {
    uplevel [list genTaskInstances $cpath \
		 $path "training testing targets" all $force] ;
}
@}

\section{Segmenting the Prototask File: splitPrototask}	
\label{sec:splitPrototask}
To create the task data file, we have to segment the prototask file
into cases that will be used for training, cases that will be used for
testing, attributes that will be used as inputs, and attributes that
will be used as targets. The \texttt{splitPrototask} command does all
of this.

\subsection{Description}
\begin{changebar}
\usage{splitPrototask}{inFile nInputs nTargets nTestCases nTrainingCases maxInstances scheme}
\end{changebar}
\texttt{SplitPrototask} splits up the prototask file into several
files. The files are put in the \texttt{/tmp} directory and given
(hopefully) unique names. The arguments to the procedure are:
\texttt{inFile}, the name of the input prototask file;
\texttt{nInputs}, and \texttt{nTargets}, the number of input
attributes and target attributes respectively; \texttt{nTestCases},
the number of testing cases; \texttt{nTrainingCases},
\begin{changebar}
number of cases for each training instance; the maximum number of
instances to generate;
\end{changebar}
\begin{changebar}
and, finally, the scheme to use when segmenting the test cases. If
\texttt{scheme} is \texttt{common}, all test sets will contain the
same cases.  If \texttt{scheme} is \texttt{hierarchical}, the test
cases will be segmented such that each instance has a unique subset of
them (just as for the training cases).
\end{changebar}

The procedure returns three lists. The first list contains the names
of the testing input files. 
\cbdelete
The second list contains the names of the testing target files.
\cbdelete
The final list contains the names of the training files (that hold
both inputs and targets). 

\subsection{Implementation}
The outline of the procedure is as follows:

\begin{changebar}
@o taskGen.tcl @{
proc splitPrototask { inFile numInputs numTargets 
		      numTestCases numTrainingCases maxInstances scheme} {

    @<build a list of the number of cases in each of the files@>
    @<split the prototask file into raw test and training files@>
    @<cut the raw test file into inputs and targets@>
    @<return the list of split and cut files@>
}
@}
\end{changebar}

The first job in \texttt{splitPrototask} is to segment the prototask
file into files containing the appropriate raw cases for the testing
and training files. To do this we need to create a list of the number
of cases in each file: \texttt{caseList}. We use the unix command
\texttt{wc} to count the number of lines in the prototask file. We
initialize \texttt{caseList} to the test set size and subtract it from
the total number of cases. We then repeatedly append the training set
size to \texttt{caseList}, and subtract it from the total number of
cases until there are no more cases left. If we don't have
\textit{exactly} zero cases left, we generate an error.
\begin{changebar}
We then truncate the number of instances to be the lesser of its
current value and \texttt{maxInstances}.
\end{changebar}

\begin{changebar}
Next, we check the \texttt{scheme}. If it is \texttt{common}, our list
is complete.  If it is \texttt{hierarchical}, we have to replace the
first element in \texttt{caseList} (the number of testing cases), with
a list of numbers where we have divided the initial number by the
number of instances. If the scheme is anything else, we generate an
error.

\begin{changebar}
@d build a list of the number of cases in each of the files @{
set numCases   [exec cat $inFile | wc -l] ;
set caseList   $numTestCases ;
incr numCases -$numTestCases ;
while { $numCases > 0 } {
    lappend caseList  $numTrainingCases ;
    incr numCases    -$numTrainingCases ;
}
set numInstances	[expr { [llength $caseList] - 1 }] ;

if { $numCases != 0 } {
    error "can't divide $inFile evenly into a $numTestCases case test set and $numTrainingCases case training sets"
}

if { $numInstances > $maxInstances } {
    set numInstances $maxInstances ;
}

if { [cequal $scheme hierarchical] } {
    if { $numTestCases % $numInstances != 0 } {
	error "can't divide $numTestCases test cases into $numInstances files"
    }
    set testSetSize	[expr { $numTestCases/$numInstances }] ;
    set caseList	[eval [list lreplace $caseList 0 0] \
			     [replicate " $testSetSize" $numInstances]] ;
} elseif { ![cequal $scheme common] } {
    error "unknown test set selection scheme \"$scheme\"" ;
}
@}
\end{changebar}
\end{changebar}

Now that the \texttt{caseList} variable has been created, we use the
\texttt{splitFile} procedure to split up the prototask file. Since
these files are only temporary, they are put in the \texttt{/tmp}
directory and given (hopefully) unique names.

\begin{changebar}
If the scheme is \texttt{common}, the first file in the list returned
by \texttt{splitFile} is the raw testing file.  If the scheme is
\texttt{hierarchical}, the first \texttt{numInstances} files are.
\end{changebar}
\begin{changebar}
The remaining \texttt{numInstances} files are the raw training files.
If we truncated the number of instances, there will be extra files at
the end of the list, and we just delete them.
\end{changebar}
Since the segmenting of the files may take some time, we use
\texttt{pwrap} to print a progress message.

\begin{changebar}
\begin{changebar}
@d split the prototask file into raw test and training files @{
pwrap "  segmenting cases..." {
    set files [splitFile $inFile /tmp/training[pid] $caseList 1] ;
}

if { [cequal $scheme common] } {
    set file	[lvarpop files] ;
    loop idx 0 $numInstances {
	lappend uncutTestFiles	$file ;
    }
} else {
    loop idx 0 $numInstances {
	lappend uncutTestFiles	[lvarpop files] ;
    }
}
loop idx 0 $numInstances {
    lappend trainingFiles	[lvarpop files] ;
}
unlink -nocomplain $files ;
@}
\end{changebar}
\end{changebar}

After we have split the prototask up by cases, we want to cut the raw
testing 
\begin{changebar}
files into inputs and targets. To do this we use the \texttt{cutFile}
procedure. Once we have cut the files, we can safely remove the
original raw test files.  As with the \texttt{splitFile} procedure, we
use \texttt{pwrap} to print a progress message. 

@d cut the raw test file into inputs and targets @{
pwrap "  splitting test inputs and targets..." {
    foreach file $uncutTestFiles {
	set files [cutFile $file $file[pid] "$numInputs $numTargets" 1] ;
	lappend testFiles	[lvarpop files] ;
	lappend targetFiles	[lvarpop files] ;
    }
    unlink -nocomplain $uncutTestFiles ;
}
@}
\end{changebar}

Now that all the raw files have been generated, we can return the list
of files: test inputs, test targets, and a list of training files:

\begin{changebar}
@d return the list of split and cut files @{
return [list $testFiles $targetFiles $trainingFiles] ;
@}
\end{changebar}

\end{document}
