% ----------------------------------------------------------------------
%
% Tcl utilities used by many procedures
%
% Author: Drew van Camp (drew@@cs.toronto.edu)
%
% Copyright (c) 1995-1996 The University of Toronto.
%
% See the file "copyright" for information on usage and redistribution
% of this file, and for a DISCLAIMER OF ALL WARRANTIES.
%
% ----------------------------------------------------------------------

\documentclass{article}
\usepackage{moretext,amstex,alltt,rcsinfo}

\pagestyle{fancyplain}
\lhead[\fancyplain{}{\leftmark}]{\fancyplain{}{\thepage}}
\rhead[\fancyplain{}{\thepage}]{\fancyplain{}{\rightmark}}

\newcommand{\tcl}{\textbf{Tcl}}
\newcommand{\tclx}{\textbf{TclX}}
\newcommand{\delve}{\textbf{Delve}}
\newcommand{\usage}[2]{\begin{alltt}{\normalfont\bfseries{}Usage:} #1 \textsl{#2}\nobreak\end{alltt}\nobreak}

\begin{document}

\rcsInfo $Id: utils.w,v 1.5 1996/02/22 20:04:07 drew Exp $

\title{Utils: Standard utilities}
\author{Drew van Camp}
\maketitle
\tableofcontents

\section{Introduction}
This module defines many standard utilities that are used throughout
\delve.  The utilities fall into several classes: file handling
routines, mathematical routines, status reporting routines, and
miscellaneous routines. About the only thing they all have in common
is that they are general purpose, and are used by sevaral other
modules.

\section{File Handling Utilities}
The utilities in this section can be used to manipulate files. 

\subsection{cutFile}
\usage{cutFile}{inFile outFilePrefix list ?force?}
This procedure reads the input file \texttt{inFile} and writes it in
n-column pieces onto a set of output files.  The name of the first
file is \texttt{outFilePrefix} with \texttt{.0} appended, and so on
incremented numerically.  The number of files created, and the number
of columns in each file, is controlled by the \texttt{list}
argument. \texttt{List} should be a list of integers $i,j,\ldots$.
The first file will contain \texttt{i} columns; the second \texttt{j}
columns, etc.  Each line in \texttt{inFile} must contain exactly
enough columns to create all the cuts, otherwise an error will be
generated.  Normally \texttt{cutFile} will not over-write existing
files.  If the optional boolean value \texttt{force} is true, it will
silently over-write.  If an error occurs while cutting the files, all
generated files will be removed.  The procedure returns a list
containing the names of all the generated files.

\subsubsection*{Implementation}
The \texttt{cutFile} procedure requires you to specify the name of the
input file, the prefix for the output files, and at least one column
length. The overall outline is straight-forward:

@o utils.tcl @{
proc cutFile { inFile outPrefix list { force 0 } } {
    set fid($inFile)	[open $inFile r] ;
    @<open all the cut files, keeping a list of their names@>
    @<calculate the column ranges for each cut file@>
    @<iterate over the input file, writing columns to each cut file@>
    @<close all the cut files@>
    close $fid($inFile) ;
    return $outFiles ;
}
@}

Before going into the details of how we cut the files, lets describe a
very important point: error clean-up.  If an error occurs while
cutting a file, we must perform some cleanup before returning.  First,
we must be sure that all open file id's are closed.  Second, as
mentioned in the command summary, we must delete all the files we've
generated.

Open file id's are kept in the array \texttt{fid}.  To close the id's
we iterate over all names in the array, closing the file id stored in
the array element.  To delete generated files, we keep a list of their
names in the variable \texttt{outFiles}. If an error occurs, we unlink
all files in this list.

@d do ``cutFile'' clean-up @{
foreach name [array names fid] {
    close $fid($name) ;
}
if { [info exists outFiles] && ![lempty $outFiles] } {
    unlink -nocomplain $outFiles ;
}
@}

Now for the guts of the cutting.  Before iterating over the input
file, we must open up \textit{all} of the output files. In this way we
can iterate over the input file once, writing the columns from each
line to the appropriate files.

@d open all the cut files, keeping a list of their names @{
set outFiles 	{} ;
loop idx 0 [llength $list] {
    set outFile	$outPrefix.$idx ;
    @<open the output cut file@>
}
@}

We put the \texttt{open} for the ouput file in a \texttt{catch}
statement so we can do clean-up if it fails. The name of the opened
file is only added to the \texttt{outFiles} list after it's opened, so
the clean-up doesn't try to delete it before it's opened. We use the
\texttt{wopen} command to open the file, using the proper mode for
\texttt{force}. 

@d open the output cut file @{
if [catch {wopen $outFile $force} msg] {
    @<do ``cutFile'' clean-up@>
    global errorInfo errorCode ;
    error $msg $errorInfo $errorCode ;
}
set fid($outFile) $msg ;
lappend outFiles $outFile ;
@}

After opening all the files, we create the arrays \texttt{first} and
\texttt{last} for holding the ranges of the columns we'll be copying,
and the variable \texttt{numColumns} to hold the number of columns
each line must have. The keys for the \texttt{first} and \texttt{last}
arrays are the names of the files.

@d calculate the column ranges for each cut file @{
set numColumns   0 ;
set column	-1 ;
foreach file $outFiles {
    set split [lvarpop list] ;

    incr numColumns $split ;

    set first($file)	[expr { $column + 1 }] ;
    set last($file)	[incr column $split] ;
}
@}

The main body of the procedure entails iterating over each line in the
input file, copying over the appropriate columns.  We use
\texttt{lrange} to extract the columns, and generate an error if we
don't have the right number of them.

@d iterate over the input file, writing columns to each cut file @{
set lineNum 0 ;
while { [gets $fid($inFile) line] != -1 } {
    if { [llength $line] != $numColumns } {
	@<do ``cutFile'' clean-up@>
	error "\"$inFile\", line $lineNum has [llength $line] columns, expected $numColumns"
    }
    foreach file $outFiles {
	puts $fid($file) [lrange $line $first($file) $last($file)] ;
    }
    incr lineNum ;
}
@}

After iterating over all the lines in the input file, we simply have
to close all of the cut files.

@d close all the cut files @{
foreach file $outFiles {
    close $fid($file) ;
}
@}

The final section of the code, described in the procedure outline,
closes the input file and returns the list containing the names of all
the generated files.

\subsection{splitFile}
\usage{splitFile}{inFile outFilePrefix list ?force?}
This procedure reads the input file \texttt{inFile} and writes it in
n-line pieces onto a set of output files.  The name of the first file
is \texttt{outFilePrefix} with \texttt{.0} appended, and so on
incremented numerically.  The number of files created, and the number
of lines in each file, is controlled by the \texttt{list} argument.
\texttt{list} should be a list of integers $i,j,\ldots$.  The first
file will contain \texttt{i} lines; the second
\texttt{j} lines, etc.  \texttt{inFile} must contain exactly enough
lines to create all the splits, otherwise an error will be generated.
Normally \texttt{splitFile} will not over-write existing files.  If
the optional boolean value \texttt{force} is true, it will silently
over-write.  If an error occurs while splitting the files, all
generated files will be removed. The procedure returns a list
containing the names of all the generated files.

\subsubsection*{Implementation}
The procedure requires you to specify the name of the input file, the
prefix for the output files, and at least one split. The overall
outline is straight-forward: 

@o utils.tcl @{
proc splitFile { inFile outPrefix list { force 0 } } {
    set fid($inFile)	[open $inFile r] ;
    @<iterate over each split, keeping a list of the files we create@>
    @<make sure there are no lines left in the unsplit file@>
    close $fid($inFile) ;
    return $outFiles ;
}
@}

Before going into details of how we split a file, we must discuss how
we handle clean-up if an error occurs.  If an error occurs while
splitting a file, we must first be sure that all open file id's are
closed. File id's are kept in the array \texttt{fid}. To close the
id's we iterate over all names in the array closing the value stored
in it. We just have to be careful that once a file is closed, we
remove its entry from the array.  As mentioned in the command summary,
if \texttt{splitFile} encounters an error, we must also deletes all
the files we have generated. To do this, we keep a list of all such
files in a variable \texttt{outFiles}. If an error occurs, we unlinks
all files in this list.

@d do ``splitFile'' clean-up @{
foreach name [array names fid] {
    close $fid($name) ;
}
if { [info exists outFiles] && ![lempty $outFiles] } {
    unlink -nocomplain $outFiles ;
}
@}

Before iterating over each split, we initialize the list of generated
files, and the extension we will append to the file prefix. Inside the
loop, we open the output file, copy over the correct number of lines,
close the file, and finally increment the file extension.

@d iterate over each split, keeping a list of the files we create @{
set outFiles 	{} ;
set ext		0 ;
foreach split $list {
    set outFile	$outPrefix.$ext ;
    @<open the output split file and save the id in ``fid(\$outFile)''@>
    @<copy ``split'' lines to the output split file@>
    @<close the output split file@>
    incr ext ;
}
@}

Opening the split file must be done inside a \texttt{catch}
command. This lets us close the input file and delete any files we've
already generated if an error occurs. We append the file name to the
list of generated files after it's open so that the clean-up won't try
to delete the file if we can't even open it. The file is opened using
\texttt{wopen}, passing in \texttt{force} to determine the proper
writing mode.

@d open the output split file and save the id in ``fid(\$outFile)''@{
if [catch {wopen $outFile $force} msg] {
    @<do ``splitFile'' clean-up@>
    global errorInfo errorCode ;
    error $msg $errorInfo $errorCode ;
}
set fid($outFile) $msg ;
lappend outFiles $outFile ;
@}

When we later close it, we have to make sure that we remove the file
id from the \texttt{fid} array so we won't try to close it if an error
occurs.

@d close the output split file @{
close $fid($outFile) ;
unset fid($outFile) ;
@}

To copy the lines from one file to another we use a simple procedure:
\texttt{copyLines}. It takes three arguments: an open file id to
read from, \texttt{in}; an open file id to write to, \texttt{out}; and
the number of lines to copy \texttt{numLines}. The procedure returns
the actual number of lines copied, which may be less than
\texttt{numLines}.

@o utils.tcl @{
proc copyLines { in out numLines } {
    loop idx 0 $numLines {
	if { [gets $in line] < 0 } {
	    break ;
	}
	puts $out $line ;
    }
    return $idx ;
}
@}

If \texttt{copyLines} isn't able to copy \texttt{split} lines, we
generate an error (after first doing clean-up).

@d copy ``split'' lines to the output split file @{
if { [copyLines $fid($inFile) $fid($outFile) $split] != $split } {
    @<do ``splitFile'' clean-up@>
    error "$inFile has too few lines to create split $ext ($split lines)" ;
}
@}

Once all the splits have been generated we see if there are any lines
left in the input file. To do that, we try to read one more line using
\texttt{gets}. If it doesn't return \texttt{-1} (end of file), then we
do clean-up and generate an error.

@d make sure there are no lines left in the unsplit file @{
if { [gets $fid($inFile) line] != -1 } {
    @<do ``splitFile'' clean-up@>
    error "$inFile has lines left over after splitting" ;
}
@}

Finally, we close the input file, and return the list of generated
files.

\subsection{wopen}
\usage{wopen}{fileName ?force?}
This utility provides a consistent interface for opening files that
are to be written to.  It opens the file called \texttt{fileName}, and
returns a \tcl\ file id which can be used in future invocations of
commands such as \texttt{puts} and \texttt{close}.  If the file cannot
be opened an error will be generated.  As well, the default behaviour
is to return an error if the file exists prior to opening.  This
behaviour can be over-ridden with the optional \texttt{force}
argument.  If \texttt{force} is non-zero, any pre-existing file will
be truncated to length zero before being opened. 

\subsubsection*{Implementation}
The procedure has one required argument, the \texttt{fileName}, and
one optional argument, the overwrite mode \texttt{force}.  The mode
for opening the file is determined from \texttt{force}.  See the \tcl\
man page for \texttt{open} for details on the file opening modes.  If
\texttt{force} is true, then we will over-write existing files.  If
not, an error will occur when we try to open it. The return value from
\texttt{open} is returned by \texttt{wopen}.

@o utils.tcl @{
proc wopen { fileName { force 0 } } {
    set mode [expr { $force ? {WRONLY CREAT TRUNC} : {WRONLY CREAT EXCL} }] ;
    open $fileName $mode ;
}
@}

\section{Mathematical Utilities}
The utilities in this section can be used to perform mathematical
operations like summations, calculation of means and variances etc.

\subsection{atPrecision} 	\label{sec:atPrecision}
\usage{atPrecision}{precision script}
\tcl\ performs all mathematical operations internally using double
precision; however, one must frequently store intermediate results in
normal \tcl\ string variables. By default, \tcl\ converts double
precision numbers to strings using 6 significant digits. This
procedure allows you to do the conversion using any arbitrary
precision. The command evaluates the \tcl\ \texttt{script}, and
anywhere a double precision number is converted to a string, it uses
\texttt{precision} significant digits. If \texttt{precision} is 17,
then there will be no loss of information in the conversion.

\subsubsection*{Implementation}
The procedure has two required arguments: the \texttt{precision} for
conversions, and the \texttt{script} to evaluate. Precision of the
conversion is controlled by the \tcl\ global variable
\texttt{tcl\_precision}. If it is not set, then \tcl\ does conversion
using six significant digits. If it is set, conversions use the number
of digits specified in the variable:

@o utils.tcl @{
proc atPrecision { precision script } {
    global tcl_precision ;
    @<save old precision and set new precision@>
    @<evaluate the script@>
    @<restore the precision@>
    @<return the result of the script@>
}
@}

When setting the precision, we must be first see if the
\texttt{tcl\_precision} variable exists yet. If it does, we save its
value in a local variable.

@d save old precision and set new precision @{
if [info exists tcl_precision] {
    set old_tcl_precision $tcl_precision ;
}
set tcl_precision $precision ;
@}

We next use \texttt{catch} to evaluate the script. This lets us
restore the precision even if an error occurs.

@d evaluate the script @{
set code [catch { uplevel $script } result] ;
@}

After evaluating the script, we restore the original precision, or
unset \texttt{tcl\_precision}, if there wasn't one initially.

@d restore the precision @{
if [info exists old_tcl_precision] {
    set tcl_precision $old_tcl_precision ;
} else {
    unset tcl_precision ;
}
@}

Finally, we return the result. If the code returned by \texttt{catch}
was one, then an error occured, and we have to return the error info
and error code as well.

@d return the result of the script @{
if { $code == 1 } {
    global errorInfo errorCode ;
    return -code $code -errorinfo $errorInfo -errorcode $errorCode $result
} else {
    return -code $code $result
}
@}

\subsection{average} 	\label{sec:average}
\usage{average}{list ?k?}
This procedure returns the arithmetic mean of a list of floating point
numbers. That is, if the list contains $n$ floating point numbers
$x_i$, it calculates:
\begin{equation}
	\frac{1}{n - k}\sum_{i = 1}^n{x_i}
\end{equation}
The parameter $n$ in the denominator represents the total degrees of
freedom in the data (the number of elements being summed), and the
parameter $k$ represents the number of constraints on the data. In the
simplest case, if $k$ is not specified, it defaults to zero; however,
it is common for $k$ to be $1$ or greater. For example, when
calculating the variance of a list of numbers, you are calculating the
average deviation from the mean, and one degree of freedom is lost
because you have already estimated the mean ($k=1$).

\subsubsection*{Implementation}
The \texttt{average} procedure has one required argument,
\texttt{list}, the list of numbers to be averaged; and one optional
argument, \texttt{k}, the number of constraints on the data. If
\texttt{k} is not specified, it defaults to zero.

@o utils.tcl @{ 
proc average { list { k 0 } } {
    expr "double([join $list +])/([llength $list] - $k)" ;
}
@| average @}

Note the use of the \texttt{join} command to transform the list of
numbers into a \tcl\ expression. This is significantly faster than
iterating over the list, accumulating the sum.

\subsection{loadMatrix} \label{sec:loadMatrix}
\usage{loadMatrix}{matrixHandle file ?indexList ...?}  

The \texttt{loadMatrix} procedure reads a \texttt{file} of floating
point numbers and copies them into the \delve\ matrix with handle
\texttt{matrixHandle} (see the document describing \texttt{d\_matrix}
for more details on matrices).  Each line of the file becomes a row in
the matrix.  Lines in the file are treated as lists, and each element
in the list is put in the corresponding column of the matrix.  If the
optional \texttt{indexList} arguments are given, they are
concatenated together to create a single list.  For each line, only
elements whose indices are in that list are loaded.

The matrix must have exactly as many rows as there are lines in the
file, and columns as there are elements on a line, or elements in the
\texttt{indexList} arguments if present. All lines in the file must
have the same number of list elements.

\subsubsection*{Implementation}
The \texttt{loadMatrix} procedure has 2 required arguments: the handle
for the matrix, and the name of the file to load. As well, it has any
number of optional arguments: the lists of element indices to
load. The procedure has two branches; the first if no elements are
given, the second if they are. In both branches, it simply iterates
over the file loading the lines into the rows of the matrix. The only
difference is that in the second branch, only selected elements are
copied in.

@o utils.tcl @{ 
proc loadMatrix { matrixHandle file args } {
    lassign [d_matrix size $matrixHandle] rows cols ;

    set row	0 ;
    if [lempty $args] {
	for_file list $file {
	    set col 0 ;
	    foreach value $list {
		d_matrix entryset $matrixHandle $row $col $value
		incr col ;
	    }
	    incr row ;
	}
    } else {
	set indices [eval concat $args] ;
	for_file list $file {
	    set col 0 ;
	    foreach index $indices {
		d_matrix entryset $matrixHandle $row $col \
		    [lindex $list $index] ;
		incr col ;
	    }
	    incr row ;
	}
    }
    if { $row != $rows } {
	error "expected $rows lines in $file but got $row" ;
    }
}
@}


\subsection{median}	\label{sec:median}
\usage{median}{list}
The \texttt{median} procedure takes a list of numbers and returns the
median value from it. The median of a list of numbers is the value for
which there are an equal number of smaller and larger values. If the
values $x_j$, $j = 0,\ldots,n-1$ are sorted into ascending order, then
the formula for the median is
\begin{equation}
	x_{\mbox{med}} =\left\{	\begin{array}{ll}
			x_{(n-1)/2} & \mbox{if $n$ is odd} \\
			\frac{1}{2}(x_{n/2} + x_{n/2-1}) 
					& \mbox{if $n$ is even}
				\end{array}
			\right.
\label{eq:median}
\end{equation}

Note that if the list has an even number of elements (in which case
there is no central value), the median is the mean of the unique
\textit{two} central values.

\subsubsection*{Implementation}
The \texttt{median} procedure takes one argument, the list of numbers
to calculate the median of. It returns that value.

@o utils.tcl @{ 
proc median { list } {
    @<sort the list and calculate its length@>
    if { $len % 2 } {
	@<if the length is odd, set the median to the middle value@>
    } else {
	@<if it's even, set it to the mean of the two middle values@>
    }
    @<return the median@>
}
@| median @}

The first task is to use the built in functions \texttt{lsort} and
\texttt{llength} to sort the list and calculate its length. The
\texttt{-real} flag must be given to \texttt{lsort} to make sure
elements are compared as floating point numbers, not alphanumeric
strings.

@d sort the list and calculate its length @{
set list   [lsort -real $list] ;
set len    [llength $list] ;
@}

If the length is odd, then the index of the middle element in the list
is the length of the list divided by two. Integer arithmetic is used,
so the value of \texttt{n1} will be truncated, effectively performing
the subtraction of 1 in Equation~\ref{eq:median}.

@d if the length is odd, set the median to the middle value @{
set n1	   [expr { $len / 2 }] ;
set median [lindex $list $n1] ;
@}

If the length is even, then we must first calculate the indices of the
two middle values \texttt{n1} and \texttt{n2}. Then add the values
together and divide by two to obtain the mean.

@d if it's even, set it to the mean of the two middle values @{
set n1	   [expr { $len / 2 }] ;
set n2	   [expr { $n1 - 1 }] ;
set median [expr { 0.5*([lindex $list $n1]+[lindex $list $n2]) }] ;
@}

The final step is to simply return the result.
@d return the median @{
return $median ;
@}

\subsection{square}	\label{sec:square}
\usage{square}{x}
This procedure returns the value of $x \times x$, where $x$ is any
valid \tcl\ expression. See the \texttt{expr} man page for full
documentation of valid expressions.

\subsubsection*{Implementation}
The {\em square} procedure takes one argument, $x$, which can be any
\tcl\ expression. It then invokes the \texttt{expr} command in the
proper context (using \texttt{uplevel}) to evaluate the product of $x$
and itself. 

@o utils.tcl @{
proc square { x } {
    uplevel [list expr ($x)*($x)] ;
}
@| square @}

Note the use of the \texttt{list} command. This ensures that if
\texttt{x} contains any spaces or special characters, the
expression passed to \texttt{expr} is quoted properly; even after it
is parsed a second time by \texttt{uplevel}.

\subsection{sum}		\label{sec:sum}
\usage{sum}{list}
The \texttt{sum} procedure takes a list of numbers (either integer or
floating point) and returns the amount obtained by adding them all
together. 

\subsubsection*{Implementation}
The implementation is straight forward. There is one required
argument: \texttt{list}, the list of numbers to sum.  The
\texttt{join} command is used to transform the list of numbers into a
\tcl\ expression which is then evaluated. This is significantly faster
than iterating over the list, accumulating the sum.

@o utils.tcl @{
proc sum { list } {
    expr [join $list +] ;
}
@| sum @}

\section{Status Reporting}
When an operation takes several seconds to complete, a status report
can make the user a little more comfortable. The routines in this
section implement several utilities to make status reporting simpler.

\subsection{pwrap}
\usage{pwrap}{msg script}
This procedure is used to wrap a \tcl\ script with a notification
message. Before \texttt{script} is evaluated, the message
\texttt{msg} will be printed to \texttt{stderr} without a trailing
new-line. After \texttt{script} is evaluated, the new-line will be
printed. 

If the global variable \texttt{pwrap\_silent} exists and is non-zero,
the message will not be printed, and \texttt{script} will simply be
evaluated.

\subsubsection*{Implementation}
This simple procedure takes two arguments: \texttt{msg} the message to
print; and \texttt{script} the script to evaluate.

If the global variable \texttt{pwrap\_silent} exists and is non-zero,
then the procedure simply evaluates \texttt{script} in the proper
context using \texttt{uplevel}, and returns the result. If
\texttt{pwrap\_silent} does not exist, or is equal to zero, then the
procedure prints \texttt{msg} to standard error, evaluates the script
using uplevel, \textit{and catches the result}. \texttt{Catch} must be
used so that even if an error occurs, we print the new-line (otherwise
error reporting would look very messy). Once the new-line is printed,
we return the proper result. If the caught code equals one, then an
error occurred and we have to return the error info and error code as
well as the result.

@o utils.tcl @{
proc pwrap { msg script } {
    global pwrap_silent ;
    if { [info exists pwrap_silent] && $pwrap_silent != 0 } {
	return [uplevel $script] ;
    }

    puts -nonewline stderr $msg ;
    set code [catch { uplevel $script } result] ;
    puts stderr "" ;
    if { $code == 1 } {
	global errorCode errorInfo ;
	return -code $code -errorinfo $errorInfo -errorcode $errorCode $result
    } else {
	return -code $code $result ;
    }
}
@}

\subsection{percent}
\usage{pwrap\_percentStart}{max ?fileId?}
\usage{pwrap\_percent}{count}
\usage{pwrap\_percentEnd}{}
These procedures can be used to print progress diagnostics about how
far through a process you have come.

\texttt{pwrap\_percentStart} initializes a counter, and prints the string
``0%%'' without a trailing new-line, to the open file id
\texttt{fileId}, or \texttt{stderr} if \texttt{fileId} isn't specified.

As you progress through your process, call \texttt{pwrap\_percent},
passing in an integer \texttt{count}, less than or equal to
\texttt{max}. It will be converted to a percentage value, and the
message will be updated to reflect that value.

Once you've finished your process, call \texttt{pwrap\_percentEnd} to
clear the message from the screen.

As with the other \texttt{pwrap} routines, setting the global variable
\texttt{pwrap\_silent} to a non-zero value will suppress printing.
As well, if the file id passed to \texttt{pwrap\_percentStart} is not a
terminal, then no messages will be printed under any circumstances.

\subsubsection*{Implementation}
The \texttt{pwrap\_percentStart} procedure has one required argument,
the \texttt{max} index to use when calculating the percentage. As well
there is an optional \texttt{fileId} argument to print the messages
to. If not specified, this defaults to \texttt{stderr}.

@o utils.tcl @{
proc pwrap_percentStart { max { fileId stderr } } {
    global pc_priv ;
    global pwrap_priv ;
    @<initialize global state variables@>
    @<print the initial message@>
}
@}

To keep track of the percentages, etc. we need a global variable that
will keep the state: \texttt{pc\_priv}. We save the current
\texttt{percent}, the \texttt{max} value, and the \texttt{fileId} in
this array.

@d initialize global state variables @{
set pc_priv(percent)	0 ;
set pc_priv(max)	$max ;
set pc_priv(fileId)  	$fileId ;
@}

We only print the initial message if \texttt{pwrap\_silent} is
non-zero (or doesn't exist), and if the file id is a terminal. We can
determine this using the \tclx\ \texttt{fstat} command.

@d print the initial message @{
if { ( ![info exists pwrap_silent] || $pwrap_silent == 0 )
     && [fstat $pc_priv(fileId) tty] } {
    puts -nonewline $pc_priv(fileId) "[format {%3s %%} $pc_priv(percent)]"
}
@}

The \texttt{pwrap\_percent} procedure takes one argument: the
\texttt{count} progress (an integer less than or equal to the maximum
value).

@o utils.tcl @{
proc pwrap_percent { count } {
    global pc_priv ;
    global pwrap_priv ;
    @<calculate the percentage@>
    @<print the progress message@>
}
@}

The first task is to calculate the percentage based on \texttt{count}.
If the value is the same as the last percentage printed, just return,
there's no point in updating the message. If it has changed, then save
the value in the global array \texttt{pc\_priv}, and continue.

@d calculate the percentage @{
set percent	[expr int(100*$count/$pc_priv(max))] ;
if { $percent == $pc_priv(percent) } {
    return ;
}
set pc_priv(percent) $percent ;
@}

As with the initial message, only print if \texttt{pwrap\_silent}
doesn't exist or is zero, and if the file id is a terminal. We start
the message with the correct number of backspaces to erase any
previous message.

@d print the progress message @{
if { ( ![info exists pwrap_silent] || $pwrap_silent == 0 )
     && [fstat $pc_priv(fileId) tty] } {
    puts -nonewline $pc_priv(fileId) "\b\b\b\b\b[format {%3s %%} $pc_priv(percent)]" ;
}
@}

The \texttt{pwrap\_percentEnd} procedure takes no arguments. 
@o utils.tcl @{
proc pwrap_percentEnd {} {
    global pc_priv ;
    global pwrap_priv ;
    @<erase the last percentage message@>
}
@}

To erase the message, we first print a blank string over the current
message, then we backspace over the blank spaces, so that whatever is
printed next appears in the correct place.

@d erase the last percentage message @{
if { ( ![info exists pwrap_silent] || $pwrap_silent == 0 )
     && [fstat $pc_priv(fileId) tty] } {
    puts -nonewline $pc_priv(fileId) "\b\b\b\b\b[format {%5s} {}]" ;
    puts -nonewline $pc_priv(fileId) "\b\b\b\b\b" ;
}
@}

\subsection{pwrap\_for\_file}
\usage{pwrap\_for\_file}{msg varName fileName script}
This procedure is a combination of \texttt{pwrap} and
\texttt{for\_file}. The procedure prints \texttt{msg} to standard error
without a trailing new-line, then iterates over each line in
\texttt{fileName}, setting the variable \texttt{varName} to contain
the text from that line. It then evaluates \texttt{script}. As you
progress through the file, the message printed to standard error is
updated to show the progress.

As with \texttt{pwrap}, if the global variable \texttt{pwrap\_silent}
exists and is non-zero, no messages are printed, the procedure behaves
exactly the same as \texttt{for\_file}.

\subsubsection*{Implementation}
This procedure takes four arguments: \texttt{msg} the message to
print; \texttt{varName}, the name of the variable to hold the text for
each line; \texttt{fileName}, the name of the file to iterate over;
and, finally, \texttt{script} the script to evaluate. It simply uses
\texttt{pwrap} to wrap the more complicated code

@o utils.tcl @{
proc pwrap_for_file { msg varName fileName script } {
    @<if pwrap\_silent is non-zero, just evaluate ``for\_file''@> 
    pwrap $msg {
	upvar $varName line ;
	@<start percentage up@>
	@<iterate over file, updating the percentage message@>
	@<end the percentage@>
    }
}
@}

The first thing we do is see if the global \texttt{pwrap\_silent}
exists and is non-zero. If it does and is, then we just return the
result of \texttt{for\_file}
@d if pwrap\_silent is non-zero, just evaluate ``for\_file'' @{
global pwrap_silent ;
if { [info exists pwrap_silent] && $pwrap_silent != 0 } {
    return [uplevel for_file $varName $fileName $script] ;
}
@}

Before iterating over all of the files, we have to initialize the
percentage counter. We use the unix commands \texttt{cat} and
\texttt{wc} to figure out how many lines are in the file: the maximum
counter.

@dstart percentage up @{
set lineNum 0 ;
pwrap_percentStart [exec cat $fileName | wc -l] ;
@}

Now we simply use \texttt{for\_file} to iterate over the file,
remembering to update the percentage counter after evaluating the
script using \texttt{uplevel}. 

@d iterate over file, updating the percentage message @{
for_file line $fileName {
    uplevel $script ;
    pwrap_percent [incr lineNum] ;
}
@}

Finally, we end the percentage counter.

@d end the percentage @{
pwrap_percentEnd ;
@}

\section{Miscellaneous Utilities}
The utilities in this section are those that don't cleanly fit into
any of the other groups. They include list manipulation routines,
error reporting routines, and many others.

\subsection{lorder}
\usage{lorder}{list orderList}
This procedure takes a \texttt{list}, and returns a new list composed
of elements from the original list, ordered according to
\texttt{orderList}.  \texttt{OrderList} should contain integer indices into
\texttt{list}.  Individual indices can occur zero or more times.  If
an index in \texttt{orderList} is less than zero or greater than or
equal to the number of elements in \texttt{list}, an empty element
will be inserted in the new list.

\subsubsection*{Implementation}
This procedure is almost trivial. It takes two lists as arguments. The
first \texttt{list} contains the elements to be re-ordered, and the
second \texttt{orderList} contains the mapping to use for the
re-ordering. The procedure simply iterates over \texttt{orderList}
extracting the ordering index, and appending the corresponding element
from \texttt{list} to the result. As mentioned in the interface
description, out of range indices will result in empty elements being
inserted in the list (thanks to the behaviour of \texttt{lindex}).

@o utils.tcl @{
proc lorder { list orderList } {
    set result {} ;
    foreach index $orderList {
	lappend result [lindex $list $index] ;
    }
    return $result ;
}
@}


\end{document}
